# HelpfulFunctions.pm
#
# A module with some functions that are used by several other modules.
#
# author:  Max-Gerd Retzlaff <m.retzlaff@gmx.net>,
#          Reyk Floeter <reyk@synack.de>
# date:    $Date: 2002/03/07 01:34:46 $
# version: $Revision: 1.3 $
#
# This application is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version. 
#
# This application is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this library; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# 02111 USA.  

package HelpfulFunctions;

use Exporter;
@ISA = ('Exporter');
@EXPORT = ('mdie', 'file_as_string', 'string_to_file', 'unique_array', 'getparent',
	'get_section_name', 'reverse_array', 'getparent', 'make_section_link');

sub mdie {
    $error = $_[0];
    $message = $_[1];
    
    # die with or without error. normally you should you mdie(0, ...,
    # to avoid bouncing messages to the sender.
    if($error) {
	die($message."Invoking die().\n");
    } else {
	print($message."Invoking exit().\n");
	exit 0;
    }
}


sub file_as_string {
    my ($target_ref, $filename) = @_;
    
    if ($filename ne '' and -f $filename) {
	my $text = "";
	
	{
	    local $/ = undef;
	    open(SOURCE, "< $filename") or 
		mdie(1, "ERROR: Can't open file \"$filename\" for reading.\n");
	    $text = <SOURCE>;
	    close(SOURCE);
	}
	$$target_ref = $text;
	return 0;
    } elsif (! -f $filename) {
    	mdie(1,"Cannot read file '$filename'\n");
    }	
    
    return 1;
}

sub string_to_file {
    my ($filename, $text, $octal_mode) = @_;
    
    open(FILE, ">$filename") or 
	mdie(1, "ERROR: Can't open file \"$filename\" for writing.\n");
    print FILE $text;
    close(FILE);

    if ( defined $octal_mode ) {
	chmod($octal_mode, $filename);
    }

    return 0;
}

# returns the unique fields of an array, preserves their order
sub unique_array {
    my @array = @_;

    my %h = ();
    grep(!$h{$_}++,@array);
    @array = keys %h;
    # replace the previous line by "@array = sort keys %h;"
    # if you want a sorted list of the unique fields

    return @array;
}

sub reverse_array {
    my @array = @_;
    my $elem;
    my @ret = ();

    while (($elem = pop(@array))) {
	push @ret, $elem;
    }

    return(@ret);
}

sub getparent {
	my ($sibling) = @_;

	my @hier = split('/', $sibling);
	my $parent = "";
	if (@hier > 1) {
		for ($i=0; $i < @hier-1; $i++) {
			$parent .= $hier[$i];
			if ($i < @hier-2) {
				$parent .= "/";
			}
		}
	}
	return($parent);
}

sub get_section_name {
	my ($section) = @_;
	
	if ((my $par = getparent($section)) ne "") {
		my @sec = split('/', $section);
		return($sec[@sec-1]);
	} else {
		return($section);
	}	
}

sub make_section_link {
	my $section = $_[0];
	my %sections_p = %{$_[1]};
	my $lastlink = $_[2];
	
        my $title = '';
        my $last = '';
        my $url = '';
        my @parts;
        @parts = split("/", $section);
        @parts = reverse_array(@parts);
        while ($part = pop @parts) {
            if ($last eq '') {
		$last .= $part;
            } else {
                  $last .= "/$part";
            }
            if (@parts > 0 || $lastlink) {
                   $url .= "<a href=\"$www_dir/".$last."/index.html\">" . %sections_p->{$last}->{Display} . "</a>";
		   if (@parts > 0) {
                   	$url .= "&nbsp;<b>/</b>&nbsp;";
		   }	
            } else {
                   $url .= %sections_p->{$last}->{Display};
            }
        }
	return($url);
}

return 1; # Last executing statement in file must be
          # non-zero, to indicate succesful loading
